@@ -1,4 +1,4 @@
-#include <sysdeps.h>
+#include <bglibs/sysdeps.h>
 #include <errno.h>
 #include <signal.h>
 #include <stdio.h>
@@ -6,10 +6,10 @@
 #include <string.h>
 #include <sys/time.h>
 #include <unistd.h>
-#include <fmt/number.h>
-#include <iobuf/iobuf.h>
-#include <msg/msg.h>
-#include <str/str.h>
+#include <bglibs/fmt.h>
+#include <bglibs/iobuf.h>
+#include <bglibs/msg.h>
+#include <bglibs/str.h>
 #include "ucspi-proxy.h"
 
 const int msg_show_pid = 1;
@@ -20,7 +20,9 @@ static unsigned long bytes_server_in = 0;
 static unsigned long bytes_server_out = 0;
 int opt_verbose = 0;
 int opt_maxline = MAXLINE;
-static unsigned opt_timeout = 30;
+static unsigned opt_connect_timeout = 30;
+static long opt_data_timeout = -1;
+static const char* opt_source_addr = NULL;
 pid_t pid;
 
 int SERVER_FD = -1;
@@ -28,22 +30,29 @@ int SERVER_FD = -1;
 struct filter_node
 {
   int fd;
-  filter_fn filter;
+  filter_fn filter_block;
+  line_fn filter_line;
+  write_fn writer;
   eof_fn at_eof;
   char* name;
+  str linebuf;
   
   struct filter_node* next;
 };
 
+int nfilters = 0;
 struct filter_node* filters = 0;
 
-static bool new_filter(int fd, filter_fn filter, eof_fn at_eof)
+static bool new_filter(int fd, filter_fn block, line_fn line, write_fn writer, eof_fn at_eof)
 {
   struct filter_node* newnode = malloc(sizeof *filters);
   if(!newnode)
     return false;
+  memset(newnode, 0, sizeof *newnode);
   newnode->fd = fd;
-  newnode->filter = filter;
+  newnode->filter_block = block;
+  newnode->filter_line = line;
+  newnode->writer = writer;
   newnode->at_eof = at_eof;
   newnode->next = 0;
   if (fd == CLIENT_IN)
@@ -63,6 +72,7 @@ static bool new_filter(int fd, filter_fn filter, eof_fn at_eof)
       ptr = ptr->next;
     ptr->next = newnode;
   }
+  ++nfilters;
   return true;
 }
 
@@ -71,12 +81,28 @@ bool set_filter(int fd, filter_fn filter, eof_fn at_eof)
   struct filter_node* node;
   for (node = filters; node != 0; node = node->next) {
     if (node->fd == fd) {
-      node->filter = filter;
+      /* Leave filter_line as-is, as a signal to handle_fd_line loop below */
+      node->filter_block = filter;
       node->at_eof = at_eof;
       return true;
     }
   }
-  return new_filter(fd, filter, at_eof);
+  return new_filter(fd, filter, NULL, NULL, at_eof);
+}
+
+bool set_line_filter(int fd, line_fn filter, write_fn writer)
+{
+  struct filter_node* node;
+  for (node = filters; node != 0; node = node->next) {
+    if (node->fd == fd) {
+      node->filter_block = NULL;
+      node->filter_line = filter;
+      node->writer = writer;
+      node->at_eof = NULL;
+      return true;
+    }
+  }
+  return new_filter(fd, NULL, filter, writer, NULL);
 }
 
 bool del_filter(int fd)
@@ -89,14 +115,42 @@ bool del_filter(int fd)
 	prev->next = curr->next;
       else
 	filters = curr->next;
+      str_free(&curr->linebuf);
       free(curr->name);
       free(curr);
+      --nfilters;
       return true;
     }
   }
   return false;
 }
 
+static void handle_fd_line(struct filter_node* filter, char* buf, ssize_t rd)
+{
+  char* end;
+  while ((end = memchr(buf, LF, rd)) != NULL) {
+    ssize_t len = end++ - buf;
+    if (len > 0 && buf[len - 1] == CR)
+      --len;
+    str_catb(&filter->linebuf, buf, len);
+    filter->filter_line(&filter->linebuf);
+    if (filter->writer != NULL && filter->linebuf.len > 0) {
+      str_catb(&filter->linebuf, CRLF, 2);
+      filter->writer(filter->linebuf.s, filter->linebuf.len);
+    }
+    filter->linebuf.len = 0;
+    rd -= end - buf;
+    buf = end;
+    if (filter->filter_block != NULL) {
+      /* Shunt off remaining data if the line function switched this filter to block mode. */
+      filter->filter_line = NULL;
+      filter->filter_block(buf, rd);
+      return;
+    }
+  }
+  str_catb(&filter->linebuf, buf, rd);
+}
+
 static void handle_fd(struct filter_node* filter)
 {
   char buf[BUFSIZE+1];
@@ -120,7 +174,10 @@ static void handle_fd(struct filter_node* filter)
       bytes_client_in += rd;
     else if (filter->fd == SERVER_FD)
       bytes_server_in += rd;
-    filter->filter(buf, rd);
+    if (filter->filter_block)
+      filter->filter_block(buf, rd);
+    else
+      handle_fd_line(filter, buf, rd);
   }
 }
 
@@ -162,6 +219,14 @@ void write_server(const char* data, ssize_t size)
   retry_write(data, size, SERVER_FD, "server", &bytes_server_out);
 }
 
+void write_line(const char* data, ssize_t size, write_fn fn)
+{
+  static str linebuf;
+  str_copyb(&linebuf, data, size);
+  str_catb(&linebuf, "\r\n", 2);
+  fn(linebuf.s, linebuf.len);
+}
+
 void log_line(const char* data, ssize_t size)
 {
   ssize_t i;
@@ -203,9 +268,12 @@ void usage(const char* message)
 {
   if(message)
     msg1(message);
-  obuf_put4s(&errbuf, "usage: ", program,
-	     " [-v] [-t timeout] [host port] ", filter_usage);
-  obuf_endl(&errbuf);
+  obuf_put5s(&errbuf, "usage: ", program, " [OPTIONS] [HOST PORT] ", filter_usage, "\n"
+    "  -v           Output verbose messages\n"
+    "  -s ADDR      Source address for outbound connections\n"
+    "  -t NUM       Time out connecting after NUM seconds\n"
+    "  -T NUM       Exit after NUM seconds of inactivity\n");
+  obuf_flush(&errbuf);
   exit(1);
 }
 
@@ -221,12 +289,18 @@ static void connfail(void)
   exit(0);
 }
 
+void connect_server(const char* hostname, const char* port)
+{
+  if ((SERVER_FD = tcp_connect(hostname, port, opt_connect_timeout, opt_source_addr)) == -1)
+    connfail();
+}
+
 static void parse_args(int argc, char* argv[])
 {
   int opt;
   unsigned tmp;
   char* end;
-  while((opt = getopt(argc, argv, "vl:t:")) != EOF) {
+  while((opt = getopt(argc, argv, "vl:s:t:T:")) != EOF) {
     switch(opt) {
     case 'v':
       opt_verbose++;
@@ -237,11 +311,20 @@ static void parse_args(int argc, char* argv[])
 	usage("Invalid maximum line length");
       opt_maxline = tmp;
       break;
+    case 's':
+      opt_source_addr = optarg;
+      break;
     case 't':
       tmp = strtoul(optarg, &end, 10);
       if (tmp == 0 || *end != 0)
-	usage("Invalid timeout");
-      opt_timeout = tmp;
+	usage("Invalid connect timeout");
+      opt_connect_timeout = tmp;
+      break;
+    case 'T':
+      tmp = strtol(optarg, &end, 10);
+      if (tmp <= 0 || *end != 0)
+	usage("Invalid data timeout");
+      opt_data_timeout = tmp;
       break;
     default:
       usage("Unknown option.");
@@ -250,11 +333,8 @@ static void parse_args(int argc, char* argv[])
   }
   if (argc - optind == 0)
     SERVER_FD = 6;
-  else if (argc - optind >= 2) {
-    if ((SERVER_FD = tcp_connect(argv[optind], argv[optind+1],
-				 opt_timeout)) == -1)
-      connfail();
-  }
+  else if (argc - optind >= 2)
+    connect_server(argv[optind], argv[optind+1]);
   else
     usage("Incorrect usage");
 
@@ -264,7 +344,6 @@ static void parse_args(int argc, char* argv[])
 
 int main(int argc, char* argv[])
 {
-  fd_set fds;
   signal(SIGALRM, SIG_IGN);
   signal(SIGHUP, SIG_IGN);
   signal(SIGPIPE, SIG_IGN);
@@ -273,20 +352,23 @@ int main(int argc, char* argv[])
   pid = getpid();
   for(;;) {
     struct filter_node* filter;
-    int maxfd = -1;
-    FD_ZERO(&fds);
-    for(filter = filters; filter; filter = filter->next) {
-      int fd = filter->fd;
-      FD_SET(fd, &fds);
-      if(fd > maxfd)
-	maxfd = fd;
+    iopoll_fd fds[nfilters];
+    int i;
+    for(i = 0, filter = filters; filter; ++i, filter = filter->next) {
+      fds[i].fd = filter->fd;
+      fds[i].events = IOPOLL_READ;
+      fds[i].revents = 0;
     }
-    while(select(maxfd+1, &fds, 0, 0, 0) == -1) {
+    while ((i = iopoll(fds, nfilters, opt_data_timeout * 1000)) < 0) {
       if(errno != EINTR)
-	usage("select failed!");
+        die1sys(111, "poll failed");
     }
-    for(filter = filters; filter; filter = filter->next)
-      if(FD_ISSET(filter->fd, &fds)) {
+    if (i == 0) {
+      msg1("Timed out");
+      break;
+    }
+    for(i = 0, filter = filters; filter; ++i, filter = filter->next)
+      if(fds[i].revents) {
 	handle_fd(filter);
 	break;
       }
